from abc import ABCMeta, abstractmethod


class AsyncBatchAdapter(metaclass=ABCMeta):
    """AsyncBatchAdapter is the interface for async Casbin adapters with multiple add and remove policy functions."""

    @abstractmethod
    async def add_policies(self, sec, ptype, rules):
        """AddPolicies adds policy rules to the storage."""
        pass

    @abstractmethod
    async def remove_policies(self, sec, ptype, rules):
        """RemovePolicies removes policy rules from the storage."""
        pass



# Copyright 2021 The casbin Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from simpleeval import EvalWithCompoundTypes
import ast


class SimpleEval(EvalWithCompoundTypes):
    """Rewrite SimpleEval.
    >>> s = SimpleEval("20 + 30 - ( 10 * 5)")
    >>> s.eval()
    0
    """

    ast_parsed_value = None

    def __init__(self, expr, functions=None):
        """Create the evaluator instance.  Set up valid operators (+,-, etc)
        functions (add, random, get_val, whatever) and names."""
        super(SimpleEval, self).__init__(functions=functions)
        if expr != "":
            self.expr = expr
            self.ast_parsed_value = ast.parse(expr.strip()).body[0].value

    def eval(self, names=None):
        """evaluate an expresssion, using the operators, functions and
        names previously set up."""

        if names:
            self.names = names

        return self._eval(self.ast_parsed_value)



from .adapter import AsyncAdapter
from .adapter_filtered import AsyncFilteredAdapter
from .batch_adapter import AsyncBatchAdapter
from .file_adapter import AsyncFileAdapter
from .update_adapter import AsyncUpdateAdapter

__all__ = [
    "AsyncAdapter",
    "AsyncFilteredAdapter",
    "AsyncBatchAdapter",
    "AsyncFileAdapter",
    "AsyncUpdateAdapter",
]



import logging
from typing import Sequence

from casbin.enforcer import Enforcer
from casbin.model import Model, FastModel, fast_policy_filter, FunctionMap
from casbin.persist.adapters import FileAdapter
from casbin.util.log import configure_logging


class FastEnforcer(Enforcer):
    _cache_key_order: Sequence[int] = None

    def __init__(
        self,
        model=None,
        adapter=None,
        enable_log=False,
        logging_config: dict = None,
        cache_key_order: Sequence[int] = None,
    ):
        self._cache_key_order = cache_key_order
        super().__init__(model, adapter, enable_log, logging_config)

    def new_model(self, path="", text=""):
        """creates a model."""
        if self._cache_key_order is None:
            m = Model()
        else:
            m = FastModel(self._cache_key_order)
        if len(path) > 0:
            m.load_model(path)
        else:
            m.load_model_from_text(text)

        return m

    def enforce(self, *rvals):
        """decides whether a "subject" can access a "object" with the operation "action",
        input parameters are usually: (sub, obj, act).
        """
        if self._cache_key_order is None:
            result, _ = self.enforce_ex(*rvals)
        else:
            keys = [rvals[x] for x in self._cache_key_order]
            with fast_policy_filter(self.model.model["p"]["p"].policy, *keys):
                result, _ = self.enforce_ex(*rvals)

        return result



# Copyright 2021 The casbin Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import TestCase
import casbin
from tests.test_enforcer import get_examples
from casbin.persist.adapters import FilteredFileAdapter
from casbin.persist.adapters.filtered_file_adapter import filter_line, filter_words


class Filter:
    # P,G are strings
    P = []
    G = []


class TestFilteredFileAdapter(TestCase):
    def test_init_filtered_adapter(self):
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        self.assertFalse(e.has_policy(["admin", "domain1", "data1", "read"]))

    def test_load_filtered_policy(self):
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = ["", "domain1"]
        filter.G = ["", "", "domain1"]
        try:
            e.load_policy()
        except:
            raise RuntimeError("unexpected error in LoadFilteredPolicy")
        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertTrue(e.has_policy(["admin", "domain2", "data2", "read"]))
        try:
            e.load_filtered_policy(filter)
        except:
            raise RuntimeError("unexpected error in LoadFilteredPolicy")

        if not e.is_filtered:
            raise RuntimeError("adapter did not set the filtered flag correctly")

        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertFalse(e.has_policy(["admin", "domain2", "data2", "read"]))

        with self.assertRaises(RuntimeError):
            e.save_policy()

        with self.assertRaises(RuntimeError):
            e.get_adapter().save_policy(e.get_model())

    def test_append_filtered_policy(self):
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = ["", "domain1"]
        filter.G = ["", "", "domain1"]
        try:
            e.load_policy()
        except:
            raise RuntimeError("unexpected error in LoadFilteredPolicy")
        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertTrue(e.has_policy(["admin", "domain2", "data2", "read"]))
        try:
            e.load_filtered_policy(filter)
        except:
            raise RuntimeError("unexpected error in LoadFilteredPolicy")

        if not e.is_filtered:
            raise RuntimeError("adapter did not set the filtered flag correctly")

        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertFalse(e.has_policy(["admin", "domain2", "data2", "read"]))

        filter.P = ["", "domain2"]
        filter.G = ["", "", "domain2"]
        try:
            e.load_increment_filtered_policy(filter)
        except:
            raise RuntimeError("unexpected error in LoadFilteredPolicy")

        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertTrue(e.has_policy(["admin", "domain2", "data2", "read"]))

    def test_filtered_policy_invalid_filter(self):
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = ["", "domain1"]

        with self.assertRaises(RuntimeError):
            e.load_filtered_policy(filter)

    def test_filtered_policy_empty_filter(self):
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)

        try:
            e.load_filtered_policy(None)
        except:
            raise RuntimeError("unexpected error in LoadFilteredPolicy")

        if e.is_filtered():
            raise RuntimeError("adapter did not reset the filtered flag correctly")

        try:
            e.save_policy()
        except:
            raise RuntimeError("unexpected error in SavePolicy")

    def test_unsupported_filtered_policy(self):
        e = casbin.Enforcer(
            get_examples("rbac_with_domains_model.conf"),
            get_examples("rbac_with_domains_policy.csv"),
        )
        filter = Filter()
        filter.P = ["", "domain1"]
        filter.G = ["", "", "domain1"]
        with self.assertRaises(ValueError):
            e.load_filtered_policy(filter)

    def test_filtered_adapter_empty_filepath(self):
        adapter = FilteredFileAdapter("")
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)

        with self.assertRaises(RuntimeError):
            e.load_filtered_policy(None)

    def test_filtered_adapter_invalid_filepath(self):
        adapter = FilteredFileAdapter(get_examples("does_not_exist_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)

        with self.assertRaises(RuntimeError):
            e.load_filtered_policy(None)

    def test_empty_filter_array(self):
        """Test filter for empty array."""
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = []
        filter.G = []

        e.load_filtered_policy(filter)
        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertTrue(e.has_policy(["admin", "domain2", "data2", "read"]))

    def test_empty_string_filter(self):
        """Test the filter for all empty strings."""
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = ["", "", ""]
        filter.G = ["", "", ""]

        e.load_filtered_policy(filter)
        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertTrue(e.has_policy(["admin", "domain2", "data2", "read"]))

    def test_mixed_empty_filter(self):
        """Test the filter for mixed empty and non-empty strings."""
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = ["", "domain1", ""]
        filter.G = ["", "", "domain1"]

        e.load_filtered_policy(filter)
        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertFalse(e.has_policy(["admin", "domain2", "data2", "read"]))

    def test_nonexistent_domain_filter(self):
        """Testing the filter for a non-existent domain."""
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = ["", "domain3"]
        filter.G = ["", "", "domain3"]

        e.load_filtered_policy(filter)
        self.assertFalse(e.has_policy(["admin", "domain3", "data3", "read"]))

    def test_empty_filter_array(self):
        """Test filter for empty array."""
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = []
        filter.G = []

        try:
            e.load_filtered_policy(filter)
        except:
            raise RuntimeError("unexpected error with empty filter arrays")

        self.assertFalse(e.is_filtered(), "Adapter should not be marked as filtered with empty filters")

        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertTrue(e.has_policy(["admin", "domain2", "data2", "read"]))

    def test_empty_string_filter(self):
        """Test the filter for all empty strings."""
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = ["", "", ""]
        filter.G = ["", "", ""]

        try:
            e.load_filtered_policy(filter)
        except:
            raise RuntimeError("unexpected error with empty string filters")

        self.assertFalse(e.is_filtered(), "Adapter should not be marked as filtered with empty string filters")

        try:
            e.save_policy()
        except:
            raise RuntimeError("unexpected error in SavePolicy with empty string filters")

        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertTrue(e.has_policy(["admin", "domain2", "data2", "read"]))

    def test_mixed_empty_filter(self):
        """Test the filter for mixed empty and non-empty strings."""
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = ["", "domain1", ""]
        filter.G = ["", "", "domain1"]

        try:
            e.load_filtered_policy(filter)
        except:
            raise RuntimeError("unexpected error with mixed empty filters")

        self.assertTrue(e.is_filtered(), "Adapter should be marked as filtered")

        with self.assertRaises(RuntimeError):
            e.save_policy()

        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertFalse(e.has_policy(["admin", "domain2", "data2", "read"]))

    def test_whitespace_filter(self):
        """Test the filter for all blank characters."""
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))
        e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
        filter = Filter()
        filter.P = [" ", "  ", "\t"]
        filter.G = ["\n", " ", "  "]

        e.load_filtered_policy(filter)

        self.assertFalse(e.is_filtered())
        self.assertTrue(e.has_policy(["admin", "domain1", "data1", "read"]))
        self.assertTrue(e.has_policy(["admin", "domain2", "data2", "read"]))

    def test_filter_line_edge_cases(self):
        """Test the boundary cases of the filter_line function."""
        adapter = FilteredFileAdapter(get_examples("rbac_with_domains_policy.csv"))

        self.assertFalse(filter_line("", [[""], [""]]))

        self.assertFalse(filter_line("invalid_line", [[""], [""]]))

        self.assertFalse(filter_line("p, admin, domain1, data1, read", None))

    def test_filter_words_edge_cases(self):
        """Test the boundary cases of the filter_words function."""
        self.assertTrue(filter_words(["p"], ["filter1", "filter2"]))

        self.assertFalse(filter_words(["p", "admin", "domain1"], []))

        line = ["admin", "domain1", "data*", "read"]
        filter = ["", "", "data1", ""]
        self.assertTrue(filter_words(line, filter))

    def test_load_filtered_policy_with_comments(self):
        """Test loading filtering policies with comments."""
        import tempfile
        import shutil

        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp_file:
            with open(get_examples("rbac_with_domains_policy.csv"), "r") as source:
                shutil.copyfileobj(source, temp_file)

            temp_file.write("\n# This is a comment\np, admin, domain1, data3, read")
            temp_file.flush()

            temp_path = temp_file.name

        try:
            adapter = FilteredFileAdapter(temp_path)
            e = casbin.Enforcer(get_examples("rbac_with_domains_model.conf"), adapter)
            filter = Filter()
            filter.P = ["", "domain1"]
            filter.G = ["", "", "domain1"]

            e.load_filtered_policy(filter)
            self.assertTrue(e.has_policy(["admin", "domain1", "data3", "read"]))
        finally:
            try:
                os.unlink(temp_path)
            except OSError:
                pass



from abc import ABCMeta, abstractmethod


class AsyncAdapter(metaclass=ABCMeta):
    """The interface for async Casbin adapters."""

    @abstractmethod
    async def load_policy(self, model):
        """loads all policy rules from the storage."""
        pass

    @abstractmethod
    async def save_policy(self, model):
        """saves all policy rules to the storage."""
        pass

    @abstractmethod
    async def add_policy(self, sec, ptype, rule):
        """adds a policy rule to the storage."""
        pass

    @abstractmethod
    async def remove_policy(self, sec, ptype, rule):
        """removes a policy rule from the storage."""
        pass

    @abstractmethod
    async def remove_filtered_policy(self, sec, ptype, field_index, *field_values):
        """removes policy rules that match the filter from the storage.
        This is part of the Auto-Save feature.
        """
        pass



from abc import ABCMeta, abstractmethod


class AsyncFilteredAdapter(metaclass=ABCMeta):
    """AsyncFilteredAdapter is the interface for async Casbin adapters supporting filtered policies."""

    @abstractmethod
    async def is_filtered(self):
        """IsFiltered returns true if the loaded policy has been filtered
        Marks if the loaded policy is filtered or not
        """
        pass

    @abstractmethod
    async def load_filtered_policy(self, model, filter):
        """Loads policy rules that match the filter from the storage."""
        pass



# Copyright 2023 The casbin Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import inspect

from casbin.core_enforcer import CoreEnforcer
from casbin.model import Model, FunctionMap
from casbin.persist.adapters.asyncio import AsyncFileAdapter, AsyncAdapter


class AsyncInternalEnforcer(CoreEnforcer):
    """
    AsyncInternalEnforcer = CoreEnforcer + Async Internal API.
    """

    def init_with_file(self, model_path, policy_path):
        """initializes an enforcer with a model file and a policy file."""
        a = AsyncFileAdapter(policy_path)
        self.init_with_adapter(model_path, a)

    def init_with_model_and_adapter(self, m, adapter=None):
        """initializes an enforcer with a model and a database adapter."""

        if not isinstance(m, Model) or adapter is not None and not isinstance(adapter, AsyncAdapter):
            raise RuntimeError("Invalid parameters for enforcer.")

        self.adapter = adapter

        self.model = m
        self.model.print_model()
        self.fm = FunctionMap.load_function_map()

        self._initialize()

    async def load_policy(self):
        """async reloads the policy from file/database."""
        need_to_rebuild = False
        new_model = copy.deepcopy(self.model)
        new_model.clear_policy()

        try:
            await self.adapter.load_policy(new_model)

            new_model.sort_policies_by_subject_hierarchy()

            new_model.sort_policies_by_priority()

            new_model.print_policy()

            if self.auto_build_role_links:
                need_to_rebuild = True
                for rm in self.rm_map.values():
                    rm.clear()

                new_model.build_role_links(self.rm_map)

            self.model = new_model

        except Exception as e:
            if self.auto_build_role_links and need_to_rebuild:
                self.build_role_links()

            raise e

    async def load_filtered_policy(self, filter):
        """async reloads a filtered policy from file/database."""
        self.model.clear_policy()

        if not hasattr(self.adapter, "is_filtered"):
            raise ValueError("filtered policies are not supported by this adapter")

        await self.adapter.load_filtered_policy(self.model, filter)

        self.model.sort_policies_by_priority()

        self.init_rm_map()
        self.model.print_policy()
        if self.auto_build_role_links:
            self.build_role_links()

    async def load_increment_filtered_policy(self, filter):
        """async append a filtered policy from file/database."""
        if not hasattr(self.adapter, "is_filtered"):
            raise ValueError("filtered policies are not supported by this adapter")

        await self.adapter.load_filtered_policy(self.model, filter)
        self.model.print_policy()
        if self.auto_build_role_links:
            self.build_role_links()

    async def save_policy(self):
        if self.is_filtered():
            raise RuntimeError("cannot save a filtered policy")

        await self.adapter.save_policy(self.model)

        if self.watcher:
            update_for_save_policy = getattr(self.watcher, "update_for_save_policy", None)
            if callable(update_for_save_policy):
                if inspect.iscoroutinefunction(update_for_save_policy):
                    await update_for_save_policy(self.model)
                else:
                    update_for_save_policy(self.model)
            else:
                self.watcher.update()

    async def _add_policy(self, sec, ptype, rule):
        """async adds a rule to the current policy."""
        rule_added = self.model.add_policy(sec, ptype, rule)
        if not rule_added:
            return rule_added

        if self.adapter and self.auto_save:
            result = await self.adapter.add_policy(sec, ptype, rule)
            if result is False:
                return False

            if self.watcher and self.auto_notify_watcher:
                update_for_add_policy = getattr(self.watcher, "update_for_add_policy", None)
                if callable(update_for_add_policy):
                    if inspect.iscoroutinefunction(update_for_add_policy):
                        await update_for_add_policy(sec, ptype, rule)
                    else:
                        update_for_add_policy(sec, ptype, rule)
                else:
                    self.watcher.update()

        return rule_added

    async def _add_policies(self, sec, ptype, rules):
        """async adds rules to the current policy."""
        rules_added = self.model.add_policies(sec, ptype, rules)
        if not rules_added:
            return rules_added

        if self.adapter and self.auto_save:
            if hasattr(self.adapter, "add_policies") is False:
                return False

            result = await self.adapter.add_policies(sec, ptype, rules)
            if result is False:
                return False

            if self.watcher and self.auto_notify_watcher:
                update_for_add_policies = getattr(self.watcher, "update_for_add_policies", None)
                if callable(update_for_add_policies):
                    if inspect.iscoroutinefunction(update_for_add_policies):
                        await update_for_add_policies(sec, ptype, rules)
                    else:
                        update_for_add_policies(sec, ptype, rules)
                else:
                    self.watcher.update()

        return rules_added

    async def _update_policy(self, sec, ptype, old_rule, new_rule):
        """async updates a rule from the current policy."""
        rule_updated = self.model.update_policy(sec, ptype, old_rule, new_rule)

        if not rule_updated:
            return rule_updated

        if self.adapter and self.auto_save:
            result = await self.adapter.update_policy(sec, ptype, old_rule, new_rule)
            if result is False:
                return False

            if self.watcher and self.auto_notify_watcher:
                self.watcher.update()

        return rule_updated

    async def _update_policies(self, sec, ptype, old_rules, new_rules):
        """async updates rules from the current policy."""
        rules_updated = self.model.update_policies(sec, ptype, old_rules, new_rules)

        if not rules_updated:
            return rules_updated

        if self.adapter and self.auto_save:
            result = await self.adapter.update_policies(sec, ptype, old_rules, new_rules)
            if result is False:
                return False

            if self.watcher and self.auto_notify_watcher:
                self.watcher.update()

        return rules_updated

    async def _update_filtered_policies(self, sec, ptype, new_rules, field_index, *field_values):
        """async deletes old rules and adds new rules."""

        old_rules = self.model.get_filtered_policy(sec, ptype, field_index, *field_values)

        if self.adapter and self.auto_save:
            try:
                old_rules = await self.adapter.update_filtered_policies(
                    sec, ptype, new_rules, field_index, *field_values
                )
            except:
                pass

        if not old_rules:
            return False

        is_rule_changed = self.model.remove_policies(sec, ptype, old_rules)
        self.model.add_policies(sec, ptype, new_rules)
        is_rule_changed = is_rule_changed and len(new_rules) != 0
        if not is_rule_changed:
            return is_rule_changed
        if sec == "g":
            self.build_role_links()
        if self.watcher and self.auto_notify_watcher:
            self.watcher.update()
        return is_rule_changed

    async def _remove_policy(self, sec, ptype, rule):
        """async removes a rule from the current policy."""
        rule_removed = self.model.remove_policy(sec, ptype, rule)
        if not rule_removed:
            return rule_removed

        if self.adapter and self.auto_save:
            result = await self.adapter.remove_policy(sec, ptype, rule)
            if result is False:
                return False

            if self.watcher and self.auto_notify_watcher:
                update_for_remove_policy = getattr(self.watcher, "update_for_remove_policy", None)
                if callable(update_for_remove_policy):
                    if inspect.iscoroutinefunction(update_for_remove_policy):
                        await update_for_remove_policy(sec, ptype, rule)
                    else:
                        update_for_remove_policy(sec, ptype, rule)
                else:
                    self.watcher.update()

        return rule_removed

    async def _remove_policies(self, sec, ptype, rules):
        """async RemovePolicies removes policy rules from the model."""
        rules_removed = self.model.remove_policies(sec, ptype, rules)
        if not rules_removed:
            return rules_removed

        if self.adapter and self.auto_save:
            if hasattr(self.adapter, "remove_policies") is False:
                return False

            result = await self.adapter.remove_policies(sec, ptype, rules)
            if result is False:
                return False

            if self.watcher and self.auto_notify_watcher:
                update_for_remove_policies = getattr(self.watcher, "update_for_remove_policies", None)
                if callable(update_for_remove_policies):
                    if inspect.iscoroutinefunction(update_for_remove_policies):
                        await update_for_remove_policies(sec, ptype, rules)
                    else:
                        update_for_remove_policies(sec, ptype, rules)
                else:
                    self.watcher.update()

        return rules_removed

    async def _remove_filtered_policy(self, sec, ptype, field_index, *field_values):
        """async removes rules based on field filters from the current policy."""
        rule_removed = self.model.remove_filtered_policy(sec, ptype, field_index, *field_values)
        if not rule_removed:
            return rule_removed

        if self.adapter and self.auto_save:
            result = await self.adapter.remove_filtered_policy(sec, ptype, field_index, *field_values)
            if result is False:
                return False

            if self.watcher and self.auto_notify_watcher:
                update_for_remove_filtered_policy = getattr(self.watcher, "update_for_remove_filtered_policy", None)
                if callable(update_for_remove_filtered_policy):
                    if inspect.iscoroutinefunction(update_for_remove_filtered_policy):
                        await update_for_remove_filtered_policy(sec, ptype, field_index, *field_values)
                    else:
                        update_for_remove_filtered_policy(sec, ptype, field_index, *field_values)
                else:
                    self.watcher.update()

        return rule_removed

    async def _remove_filtered_policy_returns_effects(self, sec, ptype, field_index, *field_values):
        """async removes rules based on field filters from the current policy."""
        rule_removed = self.model.remove_filtered_policy_returns_effects(sec, ptype, field_index, *field_values)
        if len(rule_removed) == 0:
            return rule_removed

        if self.adapter and self.auto_save:
            result = await self.adapter.remove_filtered_policy(sec, ptype, field_index, *field_values)
            if result is False:
                return False

            if self.watcher and self.auto_notify_watcher:
                self.watcher.update()

        return rule_removed

    async def get_field_index(self, ptype, field):
        """gets the index of the field name."""
        return self.model.get_field_index(ptype, field)

    async def set_field_index(self, ptype, field, index):
        """sets the index of the field name."""
        assertion = self.model["p"][ptype]
        assertion.field_index_map[field] = index



# Copyright 2023 The casbin Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Index constants
ACTION_INDEX = "act"
DOMAIN_INDEX = "dom"
SUBJECT_INDEX = "sub"
OBJECT_INDEX = "obj"
PRIORITY_INDEX = "priority"

# Effect constants
ALLOW_OVERRIDE_EFFECT = "some(where (p_eft == allow))"
DENY_OVERRIDE_EFFECT = "!some(where (p_eft == deny))"
ALLOW_AND_DENY_EFFECT = "some(where (p_eft == allow)) && !some(where (p_eft == deny))"
PRIORITY_EFFECT = "priority(p_eft) || deny"
SUBJECT_PRIORITY_EFFECT = "subjectPriority(p_eft) || deny"



# NOTE: this file exists as a backwards compatible alias. please directly
# use FilteredFileAdapter from `casbin.persist.adapters.filtered_file_adapter` instead.

from .filtered_file_adapter import Filter
from .filtered_file_adapter import FilteredFileAdapter as FilteredAdapter

__all__ = ["Filter", "FilteredAdapter"]



from abc import ABCMeta, abstractmethod


class AsyncUpdateAdapter(metaclass=ABCMeta):
    """AsyncUpdateAdapter is the interface for async Casbin adapters with add update policy function."""

    @abstractmethod
    async def update_policy(self, sec, ptype, old_rule, new_policy):
        """
        update_policy updates a policy rule from storage.
        This is part of the Auto-Save feature.
        """
        pass

    @abstractmethod
    async def update_policies(self, sec, ptype, old_rules, new_rules):
        """
        UpdatePolicies updates some policy rules to storage, like db, redis.
        """
        pass

    @abstractmethod
    async def update_filtered_policies(self, sec, ptype, new_rules, field_index, *field_values):
        """
        update_filtered_policies deletes old rules and adds new rules.
        """
        pass
